@@ -4,23 +4,24 @@ import pyntelope
 from loguru import logger
 import requests
 from collections import Counter
+import json
 
 from pyntelope import exc
 
-ATOMICASSETS_MAIN_API = "https://wax.eosusa.io/atomicassets/v1"
-ATOMICASSETS_TEST_API = "https://test.wax.eosusa.io/atomicassets/v1"
+ATOMICASSETS_MAIN_API = "https://wax.eosusa.io"
+ATOMICASSETS_TEST_API = "https://test.wax.eosusa.io"
 TIMEOUT = 3
 RETRIES = 2
 
 
 class AssetSender:
     def __init__(
-            self,
-            collection: str,
-            collection_wallet: str,
-            private_key: str,
-            api_endpoint: str = "",
-            testnet: bool = False,
+        self,
+        collection: str,
+        collection_wallet: str,
+        private_key: str,
+        api_endpoint: str = "",
+        testnet: bool = False,
     ):
         """
         Constructor
@@ -35,16 +36,22 @@ class AssetSender:
         self.private_key = private_key
         self.testnet = testnet
         # Set the default API endpoint
-        if self.testnet and not api_endpoint:
-            api_endpoint = ATOMICASSETS_TEST_API
-        elif not self.testnet and not api_endpoint:
+        if not self.testnet and not api_endpoint:
             api_endpoint = ATOMICASSETS_MAIN_API
-        self.endpoint_assets = f"{api_endpoint}/assets"
+        elif self.testnet and not api_endpoint:
+            api_endpoint = ATOMICASSETS_TEST_API
+        self.endpoint_assets = f"{api_endpoint}/atomicassets/v1/assets"
+        # Set the API endpoint for getting transfers
+        if not self.testnet:
+            transfer_endpoint = "https://wax.eosusa.io"
+        elif self.testnet:
+            transfer_endpoint = "https://test.wax.eosusa.io"
+        self.endpoint_transfers = f"{transfer_endpoint}/v2/history/get_transaction"
 
     def _get_available_assets(
-            self,
-            schema_template_list: Iterable[Tuple[str, str]],
-            sorting_key: str = "asset_id",
+        self,
+        schema_template_list: Iterable[Tuple[str, str]],
+        sorting_key: str = "asset_id",
     ) -> list:
         """
         Make request to blockchain to get available assets in collection wallet with given template IDs
@@ -69,9 +76,28 @@ class AssetSender:
         response = requests.get(self.endpoint_assets, params=payload)
         return response.json()["data"]
 
+    def _get_right_asset_id(
+        self,
+        tx_id: str,
+    ) -> str:
+        """
+        Make request to blockchain to get asset id minted with provided tx
+        :param tx_id: id of the transaction
+        :return: asset ID found in the provided minting transaction
+        """
+        logger.info(f"Making request to blockchain to find ID of freshly minted asset.")
+        payload = {
+            "id": tx_id,
+        }
+        response = requests.get(self.endpoint_transfers, params=payload)
+        logger.debug(f"Got response: {response.json()}")
+        asset_id = response.json()["actions"][1]["act"]["data"]["asset_id"]
+        logger.debug(f"Found '{asset_id}'")
+        return asset_id
+
     @staticmethod
     def _collapse_identical_schemas_templates(
-            schema_template_list: Iterable[Tuple[str, str]]
+        schema_template_list: Iterable[Tuple[str, str]]
     ):
         """
         :param schema_template_list: list with (schema, template) tuples.
@@ -85,7 +111,7 @@ class AssetSender:
 
     @staticmethod
     def _find_assets_with_highest_mints(
-            api_response, template_id: str, quantity_requested: int = 1
+        api_response, template_id: str, quantity_requested: int = 1
     ):
         """
         Finds in collection wallet given quantity of assets with given template
@@ -115,7 +141,7 @@ class AssetSender:
         return asset_ids, quantity_to_mint
 
     def _prepare_transfer_transaction(
-            self, asset_ids: any, to: str, from_wallet: str, memo: str = ""
+        self, asset_ids: any, to: str, from_wallet: str, memo: str = ""
     ):
         """
         Sends given assets from the collection wallet to given recipient wallet
@@ -138,10 +164,7 @@ class AssetSender:
                 name="to",
                 value=pyntelope.types.Name(to),
             ),
-            pyntelope.Data(
-                name="asset_ids",
-                value=asset_ids_blk
-            ),
+            pyntelope.Data(name="asset_ids", value=asset_ids_blk),
             pyntelope.Data(
                 name="memo",
                 value=pyntelope.types.String(memo),
@@ -159,15 +182,15 @@ class AssetSender:
         return action
 
     def _prepare_mint_transaction(
-            self,
-            authorized_minter: str,
-            collection_name: str,
-            schema_name: str,
-            template_id: str,
-            new_asset_owner: str,
-            immutable_data: list = None,
-            mutable_data: list = None,
-            tokens_to_back: str = "0.00000000 WAX",
+        self,
+        authorized_minter: str,
+        collection_name: str,
+        schema_name: str,
+        template_id: str,
+        new_asset_owner: str,
+        immutable_data: list = None,
+        mutable_data: list = None,
+        tokens_to_back: str = "0.00000000 WAX",
     ):
         """
         :param authorized_minter: wallet with authorisation to mint collection assets
@@ -237,11 +260,11 @@ class AssetSender:
         )
         return action
 
-    def _send_transaction(self, action) -> str:
+    def _send_transaction(self, action):
         """
         Sends transaction into blockchain.
         :param action: Action object.
-        :return: TX ID or raise ValueError exception if TX failed.
+        :return: Tuple with asset ID(s) and TX ID/False(if TX failed).
         """
         raw_transaction = pyntelope.Transaction(actions=[action])
         logger.debug("Linking transaction to the network...")
@@ -254,25 +277,36 @@ class AssetSender:
         signed_transaction = linked_transaction.sign(key=self.private_key)
         logger.debug("Sending transaction to the blockchain...")
         resp = signed_transaction.send()
+        logger.debug(json.dumps(resp))
         try:
-            return resp["transaction_id"]
+            transaction_id = resp["transaction_id"]
         except KeyError:
-            print(resp)
             error_message = resp["error"]["details"][0]["message"]
             logger.error(error_message)
+            return "None", False
+
+        asset_id = "None"
+        if resp["processed"]["action_traces"][0]["act"]["name"] == "transfer":
+            asset_id = tuple(
+                resp["processed"]["action_traces"][0]["inline_traces"][2]["act"][
+                    "data"
+                ]["asset_ids"]
+            )
+
+        return asset_id, transaction_id
 
     def send_or_mint_assets(
-            self,
-            schema_template_list: Iterable[Tuple[str, str]],
-            wallet: str,
-            memo: str = "",
+        self,
+        schema_template_list: Iterable[Tuple[str, str]],
+        wallet: str,
+        memo: str = "",
     ) -> list:
         """
         :param schema_template_list: list or tuple of tuples containing schema names and template IDs
          e.g. [("rawmaterials", "318738"), ("magmaterials", "416529")]
         :param wallet: recipient wallet
         :param memo: transaction memo
-        :return: tuple of list with asset IDs / schema-template tuples + hash of successful transaction or False
+        :return: tuple of list with asset IDs / id-schema-template tuples + hash of successful transaction or False
         """
         if not schema_template_list:
             logger.error("Schema-template list is empty!")
@@ -316,14 +350,14 @@ class AssetSender:
         return successful_tx
 
     def send_assets(
-            self,
-            assets_to_send: any,
-            wallet: str,
-            memo: str = "",
+        self,
+        assets_to_send: Iterable[any],
+        wallet: str,
+        memo: str = "",
     ) -> tuple:
         """
         Sends assets with given IDs to the provided wallet with provided memo.
-        :param assets_to_send: can be list, tuple, int or str. E.g. ('1099788246105', 1099788246106)
+        :param assets_to_send: must be an iterable type. E.g. ('1099788246105', )
         :param wallet: blockchain wallet
         :param memo: optional field for memo of the transaction
         :return: tuple of list with asset IDs + hash of successful transaction or False
@@ -331,14 +365,14 @@ class AssetSender:
         logger.info(
             f"Going to send following assets: {assets_to_send} to the wallet '{wallet}'"
         )
-        tx_return_status = self._send_transaction(
+        asset_ids, tx_return_status = self._send_transaction(
             self._prepare_transfer_transaction(
                 assets_to_send, wallet, self.collection_wallet, memo
             )
         )
         if tx_return_status:
             logger.info(f"Successfully sent: {tx_return_status}")
-            return assets_to_send, tx_return_status
+            return asset_ids, tx_return_status
         else:
             logger.critical(
                 f"Failed to send assets: {assets_to_send} to the wallet '{wallet}'!"
@@ -346,20 +380,20 @@ class AssetSender:
             return assets_to_send, False
 
     def mint_assets(
-            self,
-            schema: str,
-            template: str,
-            wallet: str,
-            quantity: int = 1,
+        self,
+        schema: str,
+        template: str,
+        wallet: str,
+        quantity: int = 1,
     ) -> list:
         """
         Mints given number of assets with given schema and template. If gets 'duplicate transaction' error will wait
-        fot TIMEOUT seconds and do RETRIES attempts of retries.
+        fot TIMEOUT seconds and do RETRIES attempts of retries. Always return 'None' instead of asset ID.
         :param schema: self-explanatory
         :param template: self-explanatory
         :param wallet: recipient wallet
         :param quantity: how many assets of the same template needs to be minted
-        :return: list of schema-template tuples + hashes of successful transactions or False
+        :return: list of id-schema-template tuples + hashes of successful transactions or False
         """
         logger.info(
             f"Going to mint {quantity} assets with template '{template}', schema '{schema}' to the wallet '{wallet}'"
@@ -370,7 +404,7 @@ class AssetSender:
         wait_time = TIMEOUT
         while minted_quantity < quantity:
             try:
-                minting_tx = self._send_transaction(
+                asset_id, minting_tx = self._send_transaction(
                     self._prepare_mint_transaction(
                         self.collection_wallet,
                         self.collection,
@@ -392,11 +426,13 @@ class AssetSender:
                 logger.error(f"Reached {retry} retries. Giving up...")
                 minting_tx = False
             if minting_tx and minting_tx not in txs:
-                logger.info(f"Successfully minted: {minting_tx}")
+                logger.info(
+                    f"Successfully minted ({asset_id}, {schema}, {template}): {minting_tx}"
+                )
                 minted_quantity += 1
                 retry = 0
                 wait_time = TIMEOUT
-                txs.append(((schema, template), minting_tx))
+                txs.append(((asset_id, schema, template), minting_tx))
                 # Sleep in order to get rid of "duplicate transaction" error
                 time.sleep(2)
             elif not minting_tx:
@@ -406,7 +442,52 @@ class AssetSender:
                 minted_quantity += 1
                 retry = 0
                 wait_time = TIMEOUT
-                txs.append(((schema, template), False))
+                txs.append(((asset_id, schema, template), False))
                 # Sleep in order to get rid of "duplicate transaction" error
                 time.sleep(2)
         return txs
+
+    def mint_assets_and_get_ids(
+        self,
+        schema: str,
+        template: str,
+        wallet: str,
+        quantity: int = 1,
+    ) -> list:
+        """
+        Same as mint_asset() but makes separate request to blockchain to fetch the real ID of minted asset.
+        :param schema: self-explanatory
+        :param template: self-explanatory
+        :param wallet: recipient wallet
+        :param quantity: how many assets of the same template needs to be minted
+        :return: list of id-schema-template tuples + hashes of successful transactions or False
+        """
+        output_list = self.mint_assets(schema, template, wallet, quantity)
+        new_output_list = []
+        logger.info("Minting finished. Now gonna try to fetch IDs.")
+        for asset in output_list:
+            transaction_id = asset[1]
+            retry = 0
+            max_retries = RETRIES + 2
+            wait_time = TIMEOUT
+            asset_id = False
+            while retry <= max_retries:
+                time.sleep(wait_time)
+                try:
+                    asset_id = self._get_right_asset_id(transaction_id)
+                except KeyError:
+                    retry += 1
+                    wait_time += 2
+                    logger.error(
+                        f"Try nr.{retry}: didn't find the transaction on blockchain. "
+                        f"Will try again in {wait_time} seconds..."
+                    )
+                else:
+                    break
+            if asset_id:
+                new_output_list.append(
+                    ((asset_id, asset[0][1], asset[0][2]), transaction_id)
+                )
+            else:
+                raise ValueError("Couldn't fetch ID! Blockchain is too slow!")
+        return new_output_list
